

from typing import Union,List,Dict,Tuple,Any,Type,TypedDict,TypeVar
from os import PathLike
import threading
import multiprocessing as mp
import numpy as np
import zmq
from enum import IntFlag

class ConnectionRole(IntFlag): ...

class BetterThread(threading.Thread): ...
class BetterProcess(mp.Process): ...
class BetterConnection(object): ...

class ClientConnection(BetterConnection): ...
class ClientNetworking(object): ...

class ServerConnection(BetterConnection): ...
class ServerNetworking(object): ...


class ClientRequest(object): ...
def stringify_client_request(request:ClientRequest) -> str: ...
def extract_client_request(request:str) -> ClientRequest: ...
